#ifndef CHIHAYA_HPP
#define CHIHAYA_HPP

/**
 * @file chihaya.hpp
 *
 * @brief Umbrella header for all **chihaya** code.
 */

#include "validate.hpp"

/**
 * @namespace chihaya
 * @brief Namespace for all **chihaya** functions.
 */
namespace chihaya {}

#endif
