#ifndef UMAPPP_UMAPPP_HPP
#define UMAPPP_UMAPPP_HPP

/**
 * @file umappp.hpp
 * @brief Overall header for the **umappp** library.
 */

#include "Options.hpp"
#include "Status.hpp"
#include "initialize.hpp"

/**
 * @namespace umappp
 * @brief Methods for UMAP.
 */
namespace umappp {}

#endif
