#ifndef SCRAN_QC_HPP
#define SCRAN_QC_HPP

#include "per_cell_qc_metrics.hpp"
#include "choose_filter_thresholds.hpp"
#include "format_filters.hpp"
#include "find_median_mad.hpp"

#include "adt_quality_control.hpp"
#include "crispr_quality_control.hpp"
#include "rna_quality_control.hpp"

/**
 * @file scran_qc.hpp
 * @brief Simple quality control for single-cell data.
 */

/**
 * @namespace scran_qc
 * @brief Simple quality control for single-cell data.
 */
namespace scran_qc {}

#endif
