#ifndef SCRAN_VARIANCES_HPP
#define SCRAN_VARIANCES_HPP

#include "fit_variance_trend.hpp"
#include "model_gene_variances.hpp"
#include "choose_highly_variable_genes.hpp"

/**
 * @file scran_variances.hpp
 * @brief Variance modelling for single-cell expression data.
 */

/**
 * @namespace scran_variances
 * @brief Variance modelling for single-cell expression data.
 */
namespace scran_variances {}

#endif
