#ifndef IRLBA_IRLBA_HPP
#define IRLBA_IRLBA_HPP

#include "compute.hpp"
#include "parallel.hpp"

/**
 * @file irlba.hpp
 * @brief Implements IRLBA
 */

/**
 * @namespace irlba
 * @brief Implements IRLBA for approximate SVD.
 */
namespace irlba {}

#endif
