#ifndef TATAMI_HDF5_HPP
#define TATAMI_HDF5_HPP

/**
 * @file tatami_hdf5.hpp
 * @brief Create matrix representations for HDF5-backed data.
 */

/**
 * @namespace tatami_hdf5
 * @brief Representations for matrix data in HDF5 files.
 */
namespace tatami_hdf5 {}

#include "CompressedSparseMatrix.hpp"
#include "DenseMatrix.hpp"
#include "load_dense_matrix.hpp"
#include "load_compressed_sparse_matrix.hpp"
#include "write_compressed_sparse_matrix.hpp"

#endif
