#ifndef KAORI_KAORI_HPP
#define KAORI_KAORI_HPP

#include "handlers/CombinatorialBarcodesPairedEnd.hpp"
#include "handlers/CombinatorialBarcodesSingleEnd.hpp"
#include "handlers/DualBarcodesPairedEnd.hpp"
#include "handlers/DualBarcodesPairedEndWithDiagnostics.hpp"
#include "handlers/DualBarcodesSingleEnd.hpp"
#include "handlers/DualBarcodesSingleEndWithDiagnostics.hpp"
#include "handlers/RandomBarcodeSingleEnd.hpp"
#include "handlers/SingleBarcodePairedEnd.hpp"
#include "handlers/SingleBarcodeSingleEnd.hpp"
#include "process_data.hpp"

/**
 * @file kaori.hpp
 * @brief Umbrella includes for the **kaori** barcode-matching library.
 */

/**
 * @namespace kaori
 * @brief Namespace for the **kaori** barcode-matching library.
 */
namespace kaori {}

#endif
