#ifndef SCRAN_BLOCKS_HPP
#define SCRAN_BLOCKS_HPP

#include "block_weights.hpp"
#include "average_vectors.hpp"

/**
 * @file scran_blocks.hpp
 * @brief Blocking utilities for **libscran**.
 */

/**
 * @namespace scran_blocks
 * @brief Blocking utilities for **libscran**.
 */
namespace scran_blocks {}

#endif
