#ifndef SCRAN_MARKERS_HPP
#define SCRAN_MARKERS_HPP

#include "score_markers_pairwise.hpp"
#include "score_markers_summary.hpp"
#include "summarize_effects.hpp"

/**
 * @file scran_markers.hpp
 * @brief Marker detection for single-cell data.
 */

/**
 * @namespace scran_markers
 * @brief Marker detection for single-cell data.
 */
namespace scran_markers {} 

#endif
