#ifndef POWERIT_POWERIT_HPP
#define POWERIT_POWERIT_HPP

/**
 * @file powerit.hpp
 * @brief Umbrella header for power iterations.
 */

#include "simple.hpp"
#include "core.hpp"

/**
 * @namespace powerit
 * @brief Namespace for power iterations.
 */
namespace powerit {}

#endif
