#ifndef TATAMI_TILEDB_HPP
#define TATAMI_TILEDB_HPP

#include "serialize.hpp"
#include "DenseMatrix.hpp"
#include "SparseMatrix.hpp"

/**
 * @file tatami_tiledb.hpp
 * @brief Umbrella header for TileDB representations.
 */

/**
 * @brief **tatami** bindings for TileDB matrices.
 * @namespace tatami_tiledb
 */
namespace tatami_tiledb {}

#endif
