#ifndef RITSUKO_RITSUKO_HPP
#define RITSUKO_RITSUKO_HPP

#include "r_missing_value.hpp"
#include "is_date_time.hpp"
#include "find_extremes.hpp"
#include "choose_missing_placeholder.hpp"
#include "parse_version_string.hpp"

/**
 * @file ritsuko.hpp
 * @brief Umbrella header for **ritsuko**.
 */

/**
 * @namespace ritsuko
 * @brief Assorted helper functions for parsing and validation.
 */
namespace ritsuko {}

#endif
