#ifndef WEIGHTEDLOWESS_WEIGHTEDLOWESS_HPP
#define WEIGHTEDLOWESS_WEIGHTEDLOWESS_HPP

#include "compute.hpp"
#include "SortBy.hpp"
#include "Options.hpp"

/**
 * @file WeightedLowess.hpp
 *
 * @brief LOWESS implementation with various weighting schemes.
 */

/**
 * @namespace WeightedLowess
 * @brief Namespace for LOWESS functions.
 */
namespace WeightedLowess {}

#endif
