#ifndef QDTSNE_QDTSNE_HPP
#define QDTSNE_QDTSNE_HPP

/**
 * @file qdtsne.hpp
 *
 * @brief Umbrella header for all **qdtsne** files.
 */

#include "Options.hpp"
#include "initialize.hpp"
#include "Status.hpp"
#include "utils.hpp"

/**
 * @namespace qdtsne
 * @brief Quick and dirty t-SNE. 
 */
namespace qdtsne {}

#endif
