#ifndef SCRAN_AGGREGATE_HPP
#define SCRAN_AGGREGATE_HPP

#include "aggregate_across_genes.hpp"
#include "aggregate_across_cells.hpp"
#include "combine_factors.hpp"
#include "clean_factor.hpp"

/**
 * @file scran_aggregate.hpp
 * @brief Aggregate single-cell expression values.
 */

/**
 * @namespace scran_aggregate
 * @brief Aggregate single-cell expression values.
 */
namespace scran_aggregate {}

#endif
