#ifndef SCRAN_PCA_HPP
#define SCRAN_PCA_HPP

#include "simple_pca.hpp"
#include "blocked_pca.hpp"

/**
 * @file scran_pca.hpp
 * @brief Principal component analysis on single-cell data.
 */

/**
 * @namespace scran_pca
 * @brief Principal component analysis on single-cell data.
 */
namespace scran_pca {}

#endif
