#ifndef SCRAN_NORM_HPP
#define SCRAN_NORM_HPP

#include "center_size_factors.hpp"
#include "choose_pseudo_count.hpp"
#include "normalize_counts.hpp"

/**
 * @file scran_norm.hpp
 * @brief Scaling normalization of single-cell data.
 */

/**
 * @namespace scran_norm
 * @brief Scaling normalization of single-cell data.
 */
namespace scran_norm {}

#endif
