#ifndef gsdECON_HPP
#define gsdECON_HPP

#include "compute.hpp"
#include "blocked.hpp"
#include "Results.hpp"
#include "Options.hpp"

/**
 * @file gsdecon.hpp
 * @brief Umbrella header for **gsdecon**.
 */

/**
 * @namespace gsdecon
 * @brief Gene set scoring with **gsdecon**.
 */
namespace gsdecon {}

#endif
