#ifndef MNNCORRECT_HPP
#define MNNCORRECT_HPP

#include "Options.hpp"
#include "ReferencePolicy.hpp"
#include "compute.hpp"

/**
 * @file mnncorrect.hpp
 * @brief Batch correction with mutual nearest neighbors.
 */

/**
 * @namespace mnncorrect
 * @brief Batch correction with mutual nearest neighbors.
 */
namespace mnncorrect {}

#endif
