#ifndef RDS2CPP_RDS2CPP_HPP
#define RDS2CPP_RDS2CPP_HPP

/**
 * @file rds2cpp.hpp
 * @brief Entrypoint header.
 */

/**
 * @brief Parse an RDS file in C++.
 */
namespace rds2cpp {}

#include "parse_rds.hpp"
#include "write_rds.hpp"

#endif
