#ifndef TAKANE_TAKANE_HPP
#define TAKANE_TAKANE_HPP

#include "_validate.hpp"
#include "_height.hpp"
#include "_satisfies_interface.hpp"
#include "_derived_from.hpp"
#include "_dimensions.hpp"

/**
 * @file takane.hpp
 * @brief Umbrella header for **takane**.
 */

/**
 * @namespace takane
 * @brief **takane** validation functions.
 */
namespace takane {}

#endif
