#ifndef RAIIGRAPH_HPP 
#define RAIIGRAPH_HPP

#include "RNGScope.hpp"
#include "Vector.hpp"
#include "Matrix.hpp"
#include "Graph.hpp"

/**
 * @file raiigraph.hpp
 *
 * @brief Utilities for manipulating **igraph** data structures in C++.
 */

/**
 * @namespace raiigraph
 * @brief Utilities for manipulating **igraph** data structures in C++.
 */
namespace raiigraph {}

#endif
